"""
Overview:
    String process module, include some useful utilities for string.
"""
from .inflection import *
from .plural import *
from .template import *
from .tree import *
from .trunc import *
