"""
Overview:
    Meta information for hbutils package.
"""

#: Title of this project (should be `hbutils`).
__TITLE__ = "hbutils"

#: Version of this project.
__VERSION__ = "0.9.0"

#: Short description of the project, will be included in ``setup.py``.
__DESCRIPTION__ = 'Some useful functions and classes in Python infrastructure development.'

#: Author of this project.
__AUTHOR__ = "HansBug"

#: Email of the authors'.
__AUTHOR_EMAIL__ = "hansbug@buaa.edu.cn"
